! This file is part of FD4: Four-Dimensional Distributed Dynamic Data structures
!  
! Copyright 2010 Matthias Lieber, ZIH, TU Dresden, Germany
! matthias.lieber@tu-dresden.de
!  
! FD4 is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!  
! FD4 is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!  
! You should have received a copy of the GNU General Public License
! along with FD4.  If not, see <http://www.gnu.org/licenses/>.

!--------------------------------------------------------------!
!                                                              !
!                     NetCDF4 Parallel I/O                     !
!                                                              !
!--------------------------------------------------------------!

#include "fd4.h"

!! Write data of an $fd4_domain to a NetCDF4/HDF5 file.
!!
!! This implementation is very simple and not optimized for performance!
!!
!! *TODO*:
!!  - open/close file for each $fd4_netcdf4_write call (can be opened before normal program termination)
!!  - how to visualize 5D data? (bin, x, y, z, time)
!!  - support for multiple bin dimensions (i.e. vars with different number of bins)
!!    - currently, one bin dimension with max size is used
!!  - optional attributes (dimension names, grid cell sizes, lat/lon/level, etc.)
!!  - better handling of face variables?
!!    - currently, interpolation to center or neglecting additional row is possible
!!  - let user choose between 'single file' mode and 'file for each output step' mode
!!  - more error/arguments checking
!!
!! @author Matthias Lieber
module fd4_netcdf4_mod

  use fd4_globaldef_mod
  use fd4_vartab_mod
  use fd4_domain_mod
  use fd4_iter_mod
  use netcdf
  
  USE_MPI
  
  implicit none
  private
  
  INCLUDE_MPIFH
   
  ! public methods
  public :: fd4_netcdf4_open
  public :: fd4_netcdf4_write
  public :: fd4_netcdf4_close
  
  public :: fd4_netcdf4_comm, fd4_domain
  
  !! FD4 NetCDF4 I/O Communicator
  type fd4_netcdf4_comm
    type(fd4_domain), pointer :: domain => null()  !! the domain which owns the blocks to iterate
    integer :: ncid                                !! NetCDF file handle
    integer :: dimids(0:4)                         !! NetCDF dimension IDs
    integer, pointer :: varid(:)                   !! NetCDF variable IDs
    integer(i_k) :: rec                            !! current output record (output time step)
    integer(i_k), pointer :: idx(:)                !! field indices of the data fields to write
    integer(i_k), pointer :: st(:)                 !! time step indices of the data fields to write
    logical :: interpolate                         !! interpolate face vars to center?
    integer(i_k) :: maxnbins                       !! max number of bins
  end type fd4_netcdf4_comm
  

contains


  !! Open a NetCDF4 file.
  !!
  subroutine fd4_netcdf4_open(nfcomm, domain, filename,  nfields, idx, st, err, int_opt)
    type(fd4_netcdf4_comm),   intent(inout) :: nfcomm         !! FD4 NetCDF4 I/O Communicator
    type(fd4_domain), target, intent(in)    :: domain         !! the domain
    character(*),             intent(in)    :: filename       !! name of the file to write
    integer(i_k),             intent(in)    :: nfields        !! number of data fields to write
    integer(i_k),             intent(in)    :: idx(nfields)   !! field indices of the data fields to write
    integer(i_k),             intent(in)    :: st(nfields)    !! time step indices of the data fields to write
    integer(i_k),             intent(out)   :: err            !! error status: 0...ok    
    logical,        optional, intent(in)    :: int_opt        !! interpolate face vars to center?
    
    integer(i_k) :: dext(0:3), i
    
    ! NetCDF4 variables
    integer :: stat
    character(FD4_VARNAME_LEN+8) :: varname
        
    err = -1
    
    FD4_L1_ASSERT(associated(nfcomm%domain),'NetCDF4 communicator is in use')
    nfcomm%domain => domain
    
    nfcomm%interpolate = .false.
    if(present(int_opt)) nfcomm%interpolate = int_opt
        
    ! copy idx and st (variable index and step)
    allocate(nfcomm%idx(nfields))
    nfcomm%idx = idx
    allocate(nfcomm%st(nfields))
    nfcomm%st = st
    
    ! allocate NetCDF variable id
    allocate(nfcomm%varid(nfields+2))
        
    ! Create the netCDF file. 
    stat = nf90_create_par( filename, IOR( NF90_HDF5, NF90_CLASSIC_MODEL ) , & 
                            domain%mpi%comm, MPI_INFO_NULL, nfcomm%ncid )
    err = stat - NF90_NOERR
    FD4_L1_ASSERT(err/=0,'nf90_create: '//trim(nf90_strerror(stat)))

    ! spatial domain size
    dext(1:3) = domain%dbnd(1:3,2) - domain%dbnd(1:3,1) + 1
    
    ! get max number of bins
    dext(0) = 1
    do i=1,size(domain%vartab)
      dext(0) = MAX(domain%vartab(i)%nbins,dext(0))
    end do
    nfcomm%maxnbins = dext(0)
    
    !! define the dimensions !!
    
    ! bin dimension
    stat = nf90_def_dim(nfcomm%ncid, 'bin', dext(0), nfcomm%dimids(0))
    err = stat - NF90_NOERR
    FD4_L1_ASSERT(err/=0,'nf90_def_dim: '//trim(nf90_strerror(stat)))

    ! 3 spatial dimensions
    stat = nf90_def_dim(nfcomm%ncid, 'x', dext(1), nfcomm%dimids(1))
    err = stat - NF90_NOERR
    FD4_L1_ASSERT(err/=0,'nf90_def_dim: '//trim(nf90_strerror(stat)))

    stat = nf90_def_dim(nfcomm%ncid, 'y', dext(2), nfcomm%dimids(2))
    err = stat - NF90_NOERR
    FD4_L1_ASSERT(err/=0,'nf90_def_dim: '//trim(nf90_strerror(stat)))

    stat = nf90_def_dim(nfcomm%ncid, 'z', dext(3), nfcomm%dimids(3))
    err = stat - NF90_NOERR
    FD4_L1_ASSERT(err/=0,'nf90_def_dim: '//trim(nf90_strerror(stat)))

    ! unlimited time dimension (records)
    stat = nf90_def_dim(nfcomm%ncid, 'time', NF90_UNLIMITED, nfcomm%dimids(4))
    err = stat - NF90_NOERR
    FD4_L1_ASSERT(err/=0,'nf90_def_dim: '//trim(nf90_strerror(stat)))


    ! define variables and fill values
    do i = 1, nfields
      
      varname = domain%vartab(idx(i))%name
      if(COUNT(idx == idx(i))>1) then
        ! variable idx appears multiple times in idx - append step
        varname = STRIM(varname)//'_'//i2s(st(i))
      end if
      ! define the variable, set fill value, set independent access mode
      call fd4_netcdf4_def_var(varname, domain%vartab(idx(i))%vnull, &
                               domain%vartab(idx(i))%nbins > 1, nfcomm%varid(i), err)
      if(err/=0) return

    end do

    ! additional variable for the block partitioning
    call fd4_netcdf4_def_var_int('blocks', -1, nfcomm%varid(nfields+1), err)
    if(err/=0) return
  
    ! additional variable for the block partitioning
    call fd4_netcdf4_def_var('weight', 0.0_r_k, .false., nfcomm%varid(nfields+2), err)
    if(err/=0) return
  
    ! End define mode. This tells netCDF we are done defining metadata.
    stat = nf90_enddef(nfcomm%ncid)
    err = stat - NF90_NOERR
    FD4_L1_ASSERT(err/=0,'nf90_enddef: '//trim(nf90_strerror(stat)))
    
    ! set current record to 0
    nfcomm%rec = 0

    err = 0

  contains

    !! define a real(r_k) variable with given name and vnull
    !!
    !! internal subroutine
    !!
    subroutine fd4_netcdf4_def_var(name, vnull, hasbins, varid, err)
      character(*), intent(in) :: name
      real(r_k), intent(in)    :: vnull
      logical, intent(in)      :: hasbins
      integer, intent(out)     :: varid
      integer, intent(inout)   :: err
      
      integer :: nf90_datatype, ld
      
      if(r_k == r8k) then
        nf90_datatype = NF90_DOUBLE
      elseif(r_k == r4k) then
        nf90_datatype = NF90_FLOAT
      else
        err = -1
        FD4_L1_ERROR('unknown real kind r_k')
      end if
      
      ! determine lowest dimension: 0=bins, 1=x, 2=y, 3=z, 4=time
      ld = 1    
      if(hasbins) ld = 0
    
      ! define the variable
      stat = nf90_def_var(nfcomm%ncid, name, nf90_datatype, nfcomm%dimids(ld:4), varid)
      err = stat - NF90_NOERR
      FD4_L1_ASSERT(err/=0,'nf90_def_var: '//trim(nf90_strerror(stat)))
      
      ! define the fill value for the variable
      stat = nf90_put_att(nfcomm%ncid, varid, '_FillValue',  vnull )
      !stat = nf90_def_var_fill(nfcomm%ncid, varid, 0, vnull )
      err = stat - NF90_NOERR
      FD4_L1_ASSERT(err/=0,'nf90_put_att: '//trim(nf90_strerror(stat)))

      ! set independend write mode
      stat = nf90_var_par_access(nfcomm%ncid, varid, NF90_INDEPENDENT)
      err = stat - NF90_NOERR
      FD4_L1_ASSERT(err/=0,'nf90_var_par_access: '//trim(nf90_strerror(stat)))

    end subroutine fd4_netcdf4_def_var


    !! define an integer variable with given name and vnull
    !!
    !! internal subroutine
    !!
    subroutine fd4_netcdf4_def_var_int(name, vnull, varid, err)
      character(*), intent(in) :: name
      integer(i_k), intent(in) :: vnull
      integer, intent(out)     :: varid
      integer, intent(inout)   :: err
    
      ! define the variable
      stat = nf90_def_var(nfcomm%ncid, name, NF90_INT, nfcomm%dimids(1:4), varid)
      err = stat - NF90_NOERR
      FD4_L1_ASSERT(err/=0,'nf90_def_var: '//trim(nf90_strerror(stat)))
      
      ! define the fill value for the variable
      stat = nf90_put_att(nfcomm%ncid, varid, '_FillValue',  vnull )
      !stat = nf90_def_var_fill(nfcomm%ncid, varid, 0, vnull )
      err = stat - NF90_NOERR
      FD4_L1_ASSERT(err/=0,'nf90_put_att: '//trim(nf90_strerror(stat)))

      ! set independend write mode
      stat = nf90_var_par_access(nfcomm%ncid, varid, NF90_INDEPENDENT)
      err = stat - NF90_NOERR
      FD4_L1_ASSERT(err/=0,'nf90_var_par_access: '//trim(nf90_strerror(stat)))

    end subroutine fd4_netcdf4_def_var_int

    
  end subroutine fd4_netcdf4_open
  

  !! Close NetCDF file.
  subroutine fd4_netcdf4_close(nfcomm, err)
    type(fd4_netcdf4_comm),   intent(inout) :: nfcomm      !! FD4 NetCDF4 I/O Communicator
    integer(i_k),             intent(out)   :: err         !! error status: 0...ok    
    integer :: stat
    
    err = -1
    
    stat = nf90_close(nfcomm%ncid)
    err = stat - NF90_NOERR
    FD4_L1_ASSERT(err/=0,'nf90_close: '//trim(nf90_strerror(stat)))

    nfcomm%domain => null()
    
    deallocate(nfcomm%idx)
    deallocate(nfcomm%st)
    deallocate(nfcomm%varid)
        
    ! Create the netCDF file. 

    err = 0

  end subroutine fd4_netcdf4_close


  !! Write current data of the domain to NetCDF file.
  !!
  !! The optional $st_opt parameter will overwrite the values given
  !! to $fd4_netcdf4_open.
  subroutine fd4_netcdf4_write(nfcomm, err, st_opt)
    type(fd4_netcdf4_comm), intent(inout), target :: nfcomm    !! FD4 NetCDF4 I/O Communicator
    integer(i_k),           intent(out)   :: err       !! error status: 0...ok    
    integer(i_k), optional, &
                    target, intent(in)    :: st_opt(:) !! time step indices of the data fields to write
    
    type(fd4_iter) :: iter
    integer(i_k) :: i, nbins, ld, idx
    real(r_k), pointer :: a(:,:,:,:)
    real(r_k), allocatable :: array(:,:,:,:)
    integer, allocatable :: array_int(:,:,:)
    integer(i_k), pointer :: st(:)
    
    integer :: stat, d, x, y, z
    integer :: start(0:4), count(0:4)
    
    err = -1
    
    if(present(st_opt)) then
      st => st_opt
    else
      st => nfcomm%st
    end if
    
    nfcomm%rec = nfcomm%rec + 1
    
    start(4) = nfcomm%rec
    count(4) = 1

    ! iterate over local blocks
    call fd4_iter_init(nfcomm%domain, iter)
    do while(associated(iter%cur))
      !b => iter%cur

      ! write variables
      do i = 1, size(nfcomm%idx)

        idx = nfcomm%idx(i)
        nbins = nfcomm%domain%vartab(idx)%nbins
        ! pointer to actual data field 
        a => iter%cur%fields(idx,st(i))%l

        !TODO: this is damn stupid slow
        ! - allocate all required array sizes in fd4_netcdf4_open or
        ! - try writing directly from the blocks (read manual how to make this)
        allocate(array(nbins,iter%cur%ext(1),iter%cur%ext(2),iter%cur%ext(3)))
        start(1:3) = iter%cur%bnd(1:3,1) - nfcomm%domain%dbnd(1:3,1) + 1
        count(1:3) = iter%cur%ext(1:3)
        start(0) = 1
        count(0) = nbins

        if(nfcomm%interpolate .and. nfcomm%domain%vartab(idx)%facevar/=0) then
          !! interpolate face vars to cell center
          select case (nfcomm%domain%vartab(idx)%facevar)
            case(1)
              do d=1,iter%cur%ext(1)
                array(1:nbins,d,:,:) = 0.5_r_k * ( a(1:nbins,d,:,:) + a(1:nbins,d+1,:,:) )
              end do          
            case(2)
              do d=1,iter%cur%ext(2)
                array(1:nbins,:,d,:) = 0.5_r_k * ( a(1:nbins,:,d,:) + a(1:nbins,:,d+1,:) )
              end do          
            case(3)
              do d=1,iter%cur%ext(3)
                array(1:nbins,:,:,d) = 0.5_r_k * ( a(1:nbins,:,:,d) + a(1:nbins,:,:,d+1) )
              end do          
          end select
        else
          !! no interpolation (face vars ignore the additional cell row)
          array(1:nbins,:,:,:) = a(1:nbins,1:iter%cur%ext(1),1:iter%cur%ext(2),1:iter%cur%ext(3))
        end if
        
        ld = 1
        if(nbins>1) ld = 0
        stat = nf90_put_var(nfcomm%ncid, nfcomm%varid(i), array, start(ld:4), count(ld:4) )
        err = stat - NF90_NOERR
        FD4_L1_ASSERT(err/=0,'nf90_put_var: '//trim(nf90_strerror(stat)))      

        deallocate(array)

      end do     
      
      ! write block partitioning
      allocate(array_int(iter%cur%ext(1),iter%cur%ext(2),iter%cur%ext(3)))
      array_int = nfcomm%domain%mpi%rank
      stat = nf90_put_var(nfcomm%ncid, nfcomm%varid(size(nfcomm%varid)-1), array_int, start(1:4), count(1:4) )
      err = stat - NF90_NOERR
      FD4_L1_ASSERT(err/=0,'nf90_put_var: '//trim(nf90_strerror(stat)))
      deallocate(array_int)
      
      ! write block weights per grid cell
      allocate(array(1,iter%cur%ext(1),iter%cur%ext(2),iter%cur%ext(3)))
      x = iter%cur%pos(1)
      y = iter%cur%pos(2)
      z = iter%cur%pos(3)
      array = REAL(nfcomm%domain%balance%blkweights(x,y,z), r_k) / REAL( PRODUCT(iter%cur%ext), r_k)
      stat = nf90_put_var(nfcomm%ncid, nfcomm%varid(size(nfcomm%varid)), array, start(1:4), count(1:4) )
      err = stat - NF90_NOERR
      FD4_L1_ASSERT(err/=0,'nf90_put_var: '//trim(nf90_strerror(stat)))
      deallocate(array)
      
      call fd4_iter_next(iter)
    end do ! blocks



    err = 0
    
  end subroutine fd4_netcdf4_write


end module fd4_netcdf4_mod
