! This file is part of FD4: Four-Dimensional Distributed Dynamic Data structures
!  
! Copyright 2010 Matthias Lieber, ZIH, TU Dresden, Germany
! matthias.lieber@tu-dresden.de
!  
! FD4 is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!  
! FD4 is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!  
! You should have received a copy of the GNU General Public License
! along with FD4.  If not, see <http://www.gnu.org/licenses/>.

!--------------------------------------------------------------!
!                                                              !
!                       NetCDF Serial I/O                      !
!                                                              !
!--------------------------------------------------------------!

#include "fd4.h"

!! Write data of an $fd4_domain to a NetCDF.
!!
!! This implementation is very simple and not optimized for performance!
!!
!! *TODO*:
!!  - how to visualize 5D data? (bin, x, y, z, time)
!!  - support for multiple bin dimensions (i.e. vars with different number of bins)
!!    - currently, one bin dimension with max size is used
!!  - optional attributes (dimension names, grid cell sizes, lat/lon/level, etc.)
!!  - better handling of face variables?
!!    - currently, interpolation to center or neglecting additional row is possible
!!  - let user choose between 'single file' mode and 'file for each output step' mode
!!  - more error/arguments checking
!!
!! @author Matthias Lieber
module fd4_netcdf4_mod

  use fd4_globaldef_mod
  use fd4_vartab_mod
  use fd4_domain_mod
  use fd4_util_mod
  use netcdf
  
  USE_MPI
  
  implicit none
  private
  
  INCLUDE_MPIFH
   
  ! public methods
  public :: fd4_netcdf4_open
  public :: fd4_netcdf4_write
  public :: fd4_netcdf4_close
  
  public :: fd4_netcdf4_comm, fd4_domain
  
  !! FD4 NetCDF4 I/O Communicator
  type fd4_netcdf4_comm
    type(fd4_domain), pointer :: domain => null()  !! the domain which owns the blocks to iterate
    integer :: ncid                                !! NetCDF file handle
    integer :: dimids(0:4)                         !! NetCDF dimension IDs
    integer, pointer :: varid(:)                   !! NetCDF variable IDs
    integer(i_k) :: rec                            !! current output record (output time step)
    integer(i_k), pointer :: idx(:)                !! field indices of the data fields to write
    integer(i_k), pointer :: st(:)                 !! time step indices of the data fields to write
    logical :: interpolate                         !! interpolate face vars to center?
    integer(i_k) :: maxnbins                       !! max number of bins
  end type fd4_netcdf4_comm
  

contains


  !! Open a NetCDF4 file.
  !!
  !! - $int_opt is not working here, just for compatibilty with parallel version in $fd4_netcdf4.
  !!
  subroutine fd4_netcdf4_open(nfcomm, domain, filename, nfields, idx, st, err, int_opt)
    type(fd4_netcdf4_comm),   intent(inout) :: nfcomm         !! FD4 NetCDF4 I/O Communicator
    type(fd4_domain), target, intent(in)    :: domain         !! the domain
    character(*),             intent(in)    :: filename       !! name of the file to write
    integer(i_k),             intent(in)    :: nfields        !! number of data fields to write
    integer(i_k),             intent(in)    :: idx(nfields)   !! field indices of the data fields to write
    integer(i_k),             intent(in)    :: st(nfields)    !! time step indices of the data fields to write
    integer(i_k),             intent(out)   :: err            !! error status: 0...ok    
    logical,        optional, intent(in)    :: int_opt        !! interpolate face vars to center?
    
    integer(i_k) :: dext(0:3), i
    character(FD4_VARNAME_LEN+8) :: varname
    
    ! NetCDF4 variables
    integer :: stat
        
    err = -1
    
    FD4_L1_ASSERT(associated(nfcomm%domain),'NetCDF4 communicator is in use')
    nfcomm%domain => domain

    nfcomm%interpolate = .false.
    if(present(int_opt)) nfcomm%interpolate = int_opt
        
    ! copy idx and st (variable index and step)
    allocate(nfcomm%idx(nfields))
    nfcomm%idx = idx
    allocate(nfcomm%st(nfields))
    nfcomm%st = st
    
    ! spatial domain size
    dext(1:3) = domain%dbnd(1:3,2) - domain%dbnd(1:3,1) + 1

    ! get max number of bins
    dext(0) = 1
    do i=1,size(domain%vartab)
      dext(0) = MAX(domain%vartab(i)%nbins,dext(0))
    end do
    nfcomm%maxnbins = dext(0)

    ! set current record to 0
    nfcomm%rec = 0

    if(nfcomm%domain%mpi%rank/=0) then
      err = 0
      return
    end if
    
    ! allocate NetCDF variable id
    allocate(nfcomm%varid(nfields+2))
        
    ! Create the netCDF file. 
    stat = nf90_create( filename, 0, nfcomm%ncid )
    err = stat - NF90_NOERR
    FD4_L1_ASSERT(err/=0,'nf90_create: '//trim(nf90_strerror(stat)))

    
    !! define the dimensions !!
    
    ! bin dimension
    stat = nf90_def_dim(nfcomm%ncid, 'bin', dext(0), nfcomm%dimids(0))
    err = stat - NF90_NOERR
    FD4_L1_ASSERT(err/=0,'nf90_def_dim: '//trim(nf90_strerror(stat)))

    ! 3 spatial dimensions
    stat = nf90_def_dim(nfcomm%ncid, 'x', dext(1), nfcomm%dimids(1))
    err = stat - NF90_NOERR
    FD4_L1_ASSERT(err/=0,'nf90_def_dim: '//trim(nf90_strerror(stat)))

    stat = nf90_def_dim(nfcomm%ncid, 'y', dext(2), nfcomm%dimids(2))
    err = stat - NF90_NOERR
    FD4_L1_ASSERT(err/=0,'nf90_def_dim: '//trim(nf90_strerror(stat)))

    stat = nf90_def_dim(nfcomm%ncid, 'z', dext(3), nfcomm%dimids(3))
    err = stat - NF90_NOERR
    FD4_L1_ASSERT(err/=0,'nf90_def_dim: '//trim(nf90_strerror(stat)))

    ! unlimited time dimension (records)
    stat = nf90_def_dim(nfcomm%ncid, 'time', NF90_UNLIMITED, nfcomm%dimids(4))
    err = stat - NF90_NOERR
    FD4_L1_ASSERT(err/=0,'nf90_def_dim: '//trim(nf90_strerror(stat)))


    ! define variables and fill values
    do i = 1, nfields
      
      varname = domain%vartab(idx(i))%name
      if(COUNT(idx == idx(i))>1) then
        ! variable idx appears multiple times in idx - append step
        varname = STRIM(varname)//'_'//i2s(st(i))
      end if
      ! define the variable, set fill value, set independent access mode
      call fd4_netcdf4_def_var(varname, domain%vartab(idx(i))%vnull, &
                               domain%vartab(idx(i))%nbins > 1, nfcomm%varid(i), err)
      if(err/=0) return

    end do

    ! additional variable for the block partitioning
    call fd4_netcdf4_def_var_int('blocks', -1, nfcomm%varid(nfields+1), err)
    if(err/=0) return
  
    ! additional variable for the block partitioning
    call fd4_netcdf4_def_var('weight', 0.0_r_k, .false., nfcomm%varid(nfields+2), err)
    if(err/=0) return

    ! End define mode. This tells netCDF we are done defining metadata.
    stat = nf90_enddef(nfcomm%ncid)
    err = stat - NF90_NOERR
    FD4_L1_ASSERT(err/=0,'nf90_enddef: '//trim(nf90_strerror(stat)))
    
    err = 0

  contains

    !! define a real(r_k) variable with given name and vnull
    !!
    !! internal subroutine
    !!
    subroutine fd4_netcdf4_def_var(name, vnull, hasbins, varid, err)
      character(*), intent(in) :: name
      real(r_k), intent(in)    :: vnull
      logical, intent(in)      :: hasbins
      integer, intent(out)     :: varid
      integer, intent(inout)   :: err
      
      integer :: nf90_datatype, ld
      
      if(r_k == r8k) then
        nf90_datatype = NF90_DOUBLE
      elseif(r_k == r4k) then
        nf90_datatype = NF90_FLOAT
      else
        err = -1
        FD4_L1_ERROR('unknown real kind r_k')
      end if
      
      ! determine lowest dimension: 0=bins, 1=x, 2=y, 3=z, 4=time
      ld = 1    
      if(hasbins) ld = 0
    
      ! define the variable
      stat = nf90_def_var(nfcomm%ncid, name, nf90_datatype, nfcomm%dimids(ld:4), varid)
      err = stat - NF90_NOERR
      FD4_L1_ASSERT(err/=0,'nf90_def_var: '//trim(nf90_strerror(stat)))
      
      ! define the fill value for the variable
      stat = nf90_put_att(nfcomm%ncid, varid, '_FillValue',  vnull )
      !stat = nf90_def_var_fill(nfcomm%ncid, varid, 0, vnull )
      err = stat - NF90_NOERR
      FD4_L1_ASSERT(err/=0,'nf90_put_att: '//trim(nf90_strerror(stat)))

    end subroutine fd4_netcdf4_def_var


    !! define an integer variable with given name and vnull
    !!
    !! internal subroutine
    !!
    subroutine fd4_netcdf4_def_var_int(name, vnull, varid, err)
      character(*), intent(in) :: name
      integer(i_k), intent(in) :: vnull
      integer, intent(out)     :: varid
      integer, intent(inout)   :: err
    
      ! define the variable
      stat = nf90_def_var(nfcomm%ncid, name, NF90_INT, nfcomm%dimids(1:4), varid)
      err = stat - NF90_NOERR
      FD4_L1_ASSERT(err/=0,'nf90_def_var: '//trim(nf90_strerror(stat)))
      
      ! define the fill value for the variable
      stat = nf90_put_att(nfcomm%ncid, varid, '_FillValue',  vnull )
      !stat = nf90_def_var_fill(nfcomm%ncid, varid, 0, vnull )
      err = stat - NF90_NOERR
      FD4_L1_ASSERT(err/=0,'nf90_put_att: '//trim(nf90_strerror(stat)))

    end subroutine fd4_netcdf4_def_var_int

    
  end subroutine fd4_netcdf4_open
  

  !! Close NetCDF file.
  subroutine fd4_netcdf4_close(nfcomm, err)
    type(fd4_netcdf4_comm),   intent(inout) :: nfcomm      !! FD4 NetCDF4 I/O Communicator
    integer(i_k),             intent(out)   :: err         !! error status: 0...ok    
    integer :: stat
    
    err = -1

    if(nfcomm%domain%mpi%rank/=0) then
      err = 0
      return
    end if
    
    stat = nf90_close(nfcomm%ncid)
    err = stat - NF90_NOERR
    FD4_L1_ASSERT(err/=0,'nf90_close: '//trim(nf90_strerror(stat)))

    nfcomm%domain => null()
    
    deallocate(nfcomm%idx)
    deallocate(nfcomm%st)
    deallocate(nfcomm%varid)
        
    ! Create the netCDF file. 

    err = 0

  end subroutine fd4_netcdf4_close


  !! Write current data of the domain to NetCDF file.
  !!
  !! The optional $st_opt parameter will overwrite the values given
  !! to $fd4_netcdf4_open.
  subroutine fd4_netcdf4_write(nfcomm, err, st_opt)
    type(fd4_netcdf4_comm), intent(inout) :: nfcomm    !! FD4 NetCDF4 I/O Communicator
    integer(i_k),           intent(out)   :: err       !! error status: 0...ok    
    integer(i_k), optional, intent(in)    :: st_opt(:) !! time step indices of the data fields to write
    target :: nfcomm, st_opt
    
    integer(i_k) :: i, ld, abnd(0:3,2), rank, facevar, nbins
    real(r_k), allocatable :: array(:,:,:,:)
    integer, allocatable :: array_int(:,:,:)
    integer(i_k), pointer :: st(:)
    integer :: stat
    integer :: start(0:4), count(0:4)
    integer(i_k) :: bbnd(3,2), z, y, x
    type(fd4_domain), pointer :: d
    
    err = -1
    d => nfcomm%domain
    rank = d%mpi%rank
    
    if(present(st_opt)) then
      st => st_opt
    else
      st => nfcomm%st
    end if
    
    nfcomm%rec = nfcomm%rec + 1
    
    ! time / record
    start(4) = nfcomm%rec
    count(4) = 1
    
    ! space
    abnd(1:3,:) = nfcomm%domain%dbnd
    start(1:3) = 1
    count(1:3) = abnd(1:3,2) - abnd(1:3,1) + 1


    !! write data fields
    
    do i = 1, size(nfcomm%idx)

      facevar = d%vartab(nfcomm%idx(i))%facevar
      nbins = d%vartab(nfcomm%idx(i))%nbins

      abnd(0,1) = 1
      abnd(0,2) = nbins
      
      if(rank == 0) then
        allocate(array(abnd(0,1):abnd(0,2),abnd(1,1):abnd(1,2),abnd(2,1):abnd(2,2),abnd(3,1):abnd(3,2)))
      else
        allocate(array(1,1,1,1)) ! dummy
      end if

      ! get field from domain
      call fd4_util_get_array(d, nfcomm%idx(i), st(i), abnd, array, err, 0)
      FD4_L1_ASSERT(err/=0,'fd4_util_get_array failed')

      if(rank==0) then
        start(0) = 1
        count(0) = nbins
        ld = 1
        if(nbins>1) ld = 0
        !print *,start(ld:4)
        !print *,count(ld:4)
        stat = nf90_put_var(nfcomm%ncid, nfcomm%varid(i), array, start(ld:4), count(ld:4) )
        err = stat - NF90_NOERR
        FD4_L1_ASSERT(err/=0,'nf90_put_var: '//trim(nf90_strerror(stat)))
      end if
      
      
      deallocate(array)
    end do




    !! write block partitioning and block weights (per gridcell)

    if(rank==0) then
      allocate(array_int(abnd(1,1):abnd(1,2),abnd(2,1):abnd(2,2),abnd(3,1):abnd(3,2)))
      allocate(array  (1,abnd(1,1):abnd(1,2),abnd(2,1):abnd(2,2),abnd(3,1):abnd(3,2)))
      array_int = 0
      array = 0
      ! loop over all blocks of the domain (also non allocated blocks)
      do z=1,d%blnum(3)
        do y=1,d%blnum(2)
          do x=1,d%blnum(1)
             bbnd(1,:) = (/ d%d(1)%bl(x), d%d(1)%bl(x+1) - 1 /)
             bbnd(2,:) = (/ d%d(2)%bl(y), d%d(2)%bl(y+1) - 1 /)
             bbnd(3,:) = (/ d%d(3)%bl(z), d%d(3)%bl(z+1) - 1 /)
             array_int(bbnd(1,1):bbnd(1,2),bbnd(2,1):bbnd(2,2),bbnd(3,1):bbnd(3,2)) = d%blkmap(x,y,z,d%blkmap_now)
             array(1,bbnd(1,1):bbnd(1,2),bbnd(2,1):bbnd(2,2),bbnd(3,1):bbnd(3,2)) = & 
                REAL(d%balance%blkweights(x,y,z), r_k) / REAL( PRODUCT(bbnd(:,2)-bbnd(:,1)+1), r_k)
          end do
        end do
      end do
      ! write partitioning
      stat = nf90_put_var(nfcomm%ncid, nfcomm%varid(size(nfcomm%varid)-1), array_int, start(1:4), count(1:4) )
      err = stat - NF90_NOERR
      FD4_L1_ASSERT(err/=0,'nf90_put_var: '//trim(nf90_strerror(stat)))
      ! write block weights
      stat = nf90_put_var(nfcomm%ncid, nfcomm%varid(size(nfcomm%varid)), array(1,:,:,:), start(1:4), count(1:4) )
      err = stat - NF90_NOERR
      FD4_L1_ASSERT(err/=0,'nf90_put_var: '//trim(nf90_strerror(stat)))
      deallocate(array_int)
      deallocate(array)
    end if


    err = 0


    
  end subroutine fd4_netcdf4_write


end module fd4_netcdf4_mod
