! This file is part of FD4: Four-Dimensional Distributed Dynamic Data structures
!  
! Copyright 2010 Matthias Lieber, ZIH, TU Dresden, Germany
! matthias.lieber@tu-dresden.de
!  
! FD4 is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!  
! FD4 is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!  
! You should have received a copy of the GNU General Public License
! along with FD4.  If not, see <http://www.gnu.org/licenses/>.

!--------------------------------------------------------------!
!                                                              !
!                      ParMETIS partitioning                   !
!                                                              !
!--------------------------------------------------------------!

#include "fd4.h"

!! ParMETIS partitioning of an $fd4_domain.
!!
!! *TODO*:
!!  - check less-blocks-then-processes case (it may happen any time during computation!)
!!  - do not make use of $fd4_domain here (-> $blkmap)?
!!
!! *IDEAS*:
!!  - consider the interconnect speed between ranks when setting adjwgt (ranks on same nodes communicate faster)
!!    - maybe a pingpong in the initialization could help finding this out dynamically
!!  - if less-blocks-then-processes case: insert fake blocks for those processes without blocks (all weights 0)
!!
!! @see fd4_domain_mod
!! @author Matthias Lieber
module fd4_part_metis_mod

  use fd4_globaldef_mod
  use fd4_domain_mod
  use fd4_mpi_mod
  use fd4_block_mod
  use rbtree_fd4_block_mod
  use stack_mod
  
  USE_MPI
  
  implicit none
  private
  
  INCLUDE_MPIFH
  
#ifdef FD4_PART_METIS_DUMP
  integer(i_k), public :: fd4_part_metis_dump = 0
#endif  

  ! public methods
  public :: fd4_part_metis, fd4_part_metis_subset
  
contains  

  include 'framework/fd4_domain.incf'
  include 'util/stack.incf'

  !! Create a new partition using ParMETIS.
  !!
  !! If wfactor is 0, ParMETIS is called without block weights.
  !!
  !!
  subroutine fd4_part_metis(domain, all_blocks, all_blocks_len, local_blocks, local_blocks_len, ubvec, itr, wfactor)
    type(fd4_domain), intent(inout) :: domain                         !! the domain
    integer(i_k)    , intent(in)    :: all_blocks_len                 !! length of $all_blocks(3,:)
    integer(i_k)    , intent(in)    :: all_blocks(3,all_blocks_len)   !! array of positions of the blocks in the new partion
    integer(i_k)    , intent(inout) :: local_blocks(:,:)              !! array of new local partition (differs from rank to rank)
    integer(i_k)    , intent(out)   :: local_blocks_len               !! length of $local_blocks(3,:)
    real(r4k)       , intent(in)    :: ubvec(1)                       !! ParMetis ubvec parameter
    real(r4k)       , intent(in)    :: itr                            !! ParMetis itr parameter
    real(r4k)       , intent(in)    :: wfactor                        !! multiply this with block's weight before converting to int
    
    integer :: rank, nprocs, comm
    integer(i_k), allocatable :: blkidx(:,:,:), blknew(:,:,:)
    integer(i_k) :: i, pos(3), npos(3), crank, dim, dir, iv, ia, err, idx(1)
    
    ! ParMetis parameters
    integer(i4k), allocatable :: vtxdist(:), part(:), vtxcounter(:), vwgt(:), adjwgt(:), vsize(:), part_global(:)
    type(stack) :: xadj, adjncy
    integer(i4k) :: wgtflag, numflag, ncon, nparts, options(0:3), edgecut
    real(r4k), allocatable :: tpwgts(:)
    integer(i_k) :: bext(3)
    type(fd4_block), pointer :: b
    
    rank   = domain%mpi%rank
    nprocs = domain%mpi%nprocs
    comm   = domain%mpi%comm
    
#if ( FD4_VERBOSE_LEVEL >= 1 )
      if(nprocs<2) then
        FD4_L1_ERROR('ParMetis load balancing requires at least 2 processes!')
      end if
#endif    
    
    ! back-up blkmap
    ! we modify blkmap_now in the following before writing the final partition to it
    domain%blkmap_last = domain%blkmap_now
    domain%blkmap_now  = 3 - domain%blkmap_now
    domain%blkmap(:,:,:,domain%blkmap_now) = domain%blkmap(:,:,:,domain%blkmap_last)

    ! mapping: 3D block coordinate -> index for parmetis
    allocate(blkidx(0:domain%blnum(1)+1,0:domain%blnum(2)+1,0:domain%blnum(3)+1))
    ! which block is on which rank?
    allocate(blknew(0:domain%blnum(1)+1,0:domain%blnum(2)+1,0:domain%blnum(3)+1))
    
    
    allocate(vtxdist(0:nprocs))
    vtxdist = 0
    allocate(vtxcounter(0:nprocs))

    call stack_create(xadj,domain%blocks%length+10)
    call stack_create(adjncy,6*domain%blocks%length)
    
    
    ! create a temporary block map (blknew)
    !  - rank 0 also takes the non-existing blocks which are in all_blocks
    !  - the other ranks take only their own blocks
    !  - removed blocks are not considered
    ! create vtxdist for parmetis
    blknew = -1
    do i=1,all_blocks_len
      pos = all_blocks(:,i)
      ! set blknew to 0 if block not exists in current partitioning
      crank = MAX(0, domain%blkmap(pos(1),pos(2),pos(3),domain%blkmap_now) ) 
      vtxdist(crank+1) = vtxdist(crank+1) + 1
      blknew(pos(1),pos(2),pos(3)) = crank
    end do
    
    
    ! check if a process has no blocks in all_blocks (would call parmetis without local blocks)
    ! this may happen if a process lost all its blocks due to empty-checks in balance_readjust
    do crank=0,nprocs-1
      if(vtxdist(crank+1)==0) then
        idx = MAXLOC(vtxdist(1:nprocs))
        idx(1) = idx(1) - 1  ! needed because of lbound = 0, maxloc assumes lbound = 1
        if(vtxdist(idx(1)+1)<2) exit ! exit, if proc with max vtx count has <2 vtxs
#if ( FD4_VERBOSE_LEVEL >= 2 )
        if(rank==0) then
          write(fd4_out,'(A,A)') FD4_LINEHD, &
             'Emergency block pre-migration from '//i2s(idx(1))//' to '//i2s(crank)
          write(fd4_out,'(A,A)') FD4_LINEHD, &
             '          block counts:            '//i2s(vtxdist(idx(1)+1))//'    '//i2s(vtxdist(crank+1))
        end if
#endif
        vtxdist(idx(1)+1) = vtxdist(idx(1)+1) - 1
        vtxdist(crank+1) = 1
        ! modify blkmap and blknew
        do i=1,all_blocks_len
          pos = all_blocks(:,i)
          if(blknew(pos(1),pos(2),pos(3))==idx(1)) then
            blknew(pos(1),pos(2),pos(3)) = crank
            if(domain%blkmap(pos(1),pos(2),pos(3),domain%blkmap_now)>=0) then
              ! only set blkmap if the block is already assigned
              domain%blkmap(pos(1),pos(2),pos(3),domain%blkmap_now) = crank
            end if
            exit
          end if
        end do
      end if
    end do
    
        

    ! sum vtxdist up
    do i=2,nprocs
      vtxdist(i) = vtxdist(i) + vtxdist(i-1)
    end do
    vtxcounter=vtxdist


    ! create blkidx, maps 3D block coordinates to rank-consecutive IDs used for parmetis
    blkidx = -1
    do i=1,all_blocks_len
      pos = all_blocks(:,i)
      crank = blknew(pos(1),pos(2),pos(3))
      blkidx(pos(1),pos(2),pos(3)) = vtxcounter(crank)
      vtxcounter(crank) = vtxcounter(crank) + 1
    end do    

    ! apply periodic boundary on blknew and blkidx
    pos = domain%blnum
    if(domain%periodic(1)) then
      blkidx(0,1:pos(2),1:pos(3)) = blkidx(pos(1),1:pos(2),1:pos(3))
      blkidx(pos(1)+1,1:pos(2),1:pos(3)) = blkidx(1,1:pos(2),1:pos(3))
      blknew(0,1:pos(2),1:pos(3)) = blknew(pos(1),1:pos(2),1:pos(3))
      blknew(pos(1)+1,1:pos(2),1:pos(3)) = blknew(1,1:pos(2),1:pos(3))
    end if
    if(domain%periodic(2)) then
      blkidx(1:pos(1),0,1:pos(3)) = blkidx(1:pos(1),pos(2),1:pos(3))
      blkidx(1:pos(1),pos(2)+1,1:pos(3)) = blkidx(1:pos(1),1,1:pos(3))
      blknew(1:pos(1),0,1:pos(3)) = blknew(1:pos(1),pos(2),1:pos(3))
      blknew(1:pos(1),pos(2)+1,1:pos(3)) = blknew(1:pos(1),1,1:pos(3))
    end if
    if(domain%periodic(3)) then
      blkidx(1:pos(1),1:pos(2),0) = blkidx(1:pos(1),1:pos(2),pos(3))
      blkidx(1:pos(1),1:pos(2),pos(3)+1) = blkidx(1:pos(1),1:pos(2),1)
      blknew(1:pos(1),1:pos(2),0) = blknew(1:pos(1),1:pos(2),pos(3))
      blknew(1:pos(1),1:pos(2),pos(3)+1) = blknew(1:pos(1),1:pos(2),1)
    end if      
    
#ifdef FD4_BALANCE_METIS_DEBUG    
    write(100+rank,*) '== FD4_BALANCE_METIS_DEBUG =='
    do iv=1,domain%blnum(3)
      do ia=1,domain%blnum(2)
        write(100+rank,*) blkidx(1:domain%blnum(1),ia,iv)
      end do
      write(100+rank,*) '---'
    end do
    do iv=1,domain%blnum(3)
      do ia=1,domain%blnum(2)
        write(100+rank,*) blknew(1:domain%blnum(1),ia,iv)
      end do
      write(100+rank,*) '---'
    end do
#endif    

    
    ! Create xadj and adjncy for parmetis
    !
    call stack_push(xadj,0)    
    do i=1,all_blocks_len
      pos = all_blocks(:,i)
      if( blknew(pos(1),pos(2),pos(3)) == rank ) then
        do dim=1,3
          do dir=1,2
            npos(:) = pos(:)
            npos(dim) = npos(dim) + fd4_incdec(dir)
            ! put all neighbors to adjcny
            if(blkidx(npos(1),npos(2),npos(3))>=0) then
              call stack_push(adjncy,blkidx(npos(1),npos(2),npos(3)))
            end if
          end do
        end do            
        call stack_push(xadj,adjncy%pos)
      end if
    end do
    

#ifdef FD4_BALANCE_METIS_DEBUG    
    write(100+rank,*) 'xadj ',0,' - ',xadj%pos-1
    write(100+rank,*) xadj%a(1:xadj%pos)
    write(100+rank,*) 'adjncy ',0,' - ',adjncy%pos-1
    write(100+rank,*) adjncy%a(1:adjncy%pos)
    write(100+rank,*) 'vtxdist ',lbound(vtxdist),' - ',ubound(vtxdist)
    write(100+rank,*) vtxdist
#endif    
    
    
    allocate(vwgt(0:xadj%pos-2))
    allocate(vsize(0:xadj%pos-2))
    allocate(adjwgt(0:adjncy%pos-1))
    vwgt=0 
    vsize=0
    adjwgt=0

    ! set vwgt, vsize, and adjwgt
    !
    do i=1,all_blocks_len
      pos = all_blocks(:,i)
      if( blknew(pos(1),pos(2),pos(3)) == rank ) then
        iv = blkidx(pos(1),pos(2),pos(3)) - vtxdist(rank)
        FD4_L3_ASSERT(iv<0,'inconsistency in calc. of adjwgt')
        FD4_L3_ASSERT(iv>xadj%pos-2,'inconsistency in calc. of adjwgt')
        if(wfactor>0.0) then
          ! USE BLOCK WEIGHTS
          ! get block extent
          do dim=1,3
            bext(dim) = domain%d(dim)%bl(pos(dim)+1) - domain%d(dim)%bl(pos(dim))
          end do
          ! set redistribution costs of the block
          if(domain%blkmap(pos(1),pos(2),pos(3),domain%blkmap_now) >= 0) vsize(iv) = PRODUCT(bext)
          ! get the block to get its comp. weight
          call fd4_domain_c2i(pos, idx(1))
          b => rbtree_fd4_block_get(domain%blocks,idx(1))
          if( associated(b) ) then
            ! block is allocated
            vwgt(iv) = INT( b%weight * wfactor, i4k)
          else
            ! block is not allocated (new created)
            vwgt(iv) = PRODUCT(bext) * domain%balance%defaultweight
          end if
        else
          ! set comp. weight of the block
          vwgt(iv) = 1
          ! set redistribution costs of the block
          if(domain%blkmap(pos(1),pos(2),pos(3),domain%blkmap_now) >= 0) vsize(iv) = 1
        end if
        ia=xadj%a(iv+1)
        do dim=1,3
          do dir=1,2
            npos(:) = pos(:)
            npos(dim) = npos(dim) + fd4_incdec(dir)
            if(blknew(npos(1),npos(2),npos(3))>=0) then
              if(wfactor>0.0) then
                ! USE BLOCK WEIGHTS
                adjwgt(ia) = domain%nghosts(dim) * PRODUCT(bext) / bext(dim)
              else
                adjwgt(ia) = 1
              endif
              ia=ia+1
            end if
          end do
        end do
        FD4_L3_ASSERT(xadj%a(iv+2)/=ia,'inconsistency in calc. of adjwgt')
      end if      
    end do    
    
    

#ifdef FD4_BALANCE_METIS_DEBUG    
    write(100+rank,*) 'vwgt ',lbound(vwgt,1),' - ',ubound(vwgt,1)
    write(100+rank,*) vwgt
    write(100+rank,*) 'vsize ',lbound(vsize,1),' - ',ubound(vsize,1)
    write(100+rank,*) vsize
    write(100+rank,*) 'adjwgt ',lbound(adjwgt,1),' - ',ubound(adjwgt,1)
    write(100+rank,*) adjwgt
#endif
    

    if(wfactor>0.0) then
      ! USE BLOCK WEIGHTS
      wgtflag = 3
    else
      wgtflag = 0
    endif

    numflag = 0 ! we are using C-Style numbering!
    ncon = 1
    nparts = nprocs
    allocate(tpwgts(nprocs))
    tpwgts = 1.0_r4k/REAL(nparts,r4k)
    !options(0:3) = (/1,127,15,1/)
    options(0:3) = 0
    allocate(part(0:xadj%pos-2))
    part = 0



#ifdef FD4_PART_METIS_DUMP
    if(fd4_part_metis_dump>=1) then
      if(rank==0) write(fd4_out,'(A,A)') FD4_LINEHD,'dumping ParMetis parameters...'
      !! dump parmetis parameters to file
      write(1000+rank,*) lbound(vtxdist,1),ubound(vtxdist,1),'vtxdist'
      write(1000+rank,*) vtxdist
      write(1000+rank,*) 0,xadj%pos-1,'xadj'
      write(1000+rank,*) xadj%a(1:xadj%pos)
      write(1000+rank,*) 0,adjncy%pos-1,'adjncy'
      write(1000+rank,*) adjncy%a(1:adjncy%pos)
      write(1000+rank,*) lbound(vwgt,1),ubound(vwgt,1),'vwgt'
      write(1000+rank,*) vwgt
      write(1000+rank,*) lbound(vsize,1),ubound(vsize,1),'vsize'
      write(1000+rank,*) vsize
      write(1000+rank,*) lbound(adjwgt,1),ubound(adjwgt,1),'adjwgt'
      write(1000+rank,*) adjwgt
      write(1000+rank,*) wgtflag
      write(1000+rank,*) numflag
      write(1000+rank,*) ncon
      write(1000+rank,*) nparts
      write(1000+rank,*) lbound(tpwgts,1),ubound(tpwgts,1),'tpwgts'
      write(1000+rank,*) tpwgts
      write(1000+rank,*) lbound(ubvec,1),ubound(ubvec,1),'ubvec'
      write(1000+rank,*) ubvec
      write(1000+rank,*) itr
      write(1000+rank,*) lbound(options,1),ubound(options,1),'options'
      write(1000+rank,*) options
      write(1000+rank,*) lbound(part,1),ubound(part,1),'part'
      write(1000+rank,*) part
      CALL MPI_Barrier(comm,err)
      if(rank==0) write(fd4_out,'(A,A)') FD4_LINEHD,'ParMetis parameters dumped.'
      !if(fd4_part_metis_dump==2) call MPI_Abort(MPI_COMM_WORLD,1,err)
      if(fd4_part_metis_dump==2) stop
      fd4_part_metis_dump = 0
    end if
#endif  


#if ( FD4_VERBOSE_LEVEL >= 4 )
      if(rank==0) then
        FD4_L4_INFO('calling ParMetis')
      end if
#endif
    
    ! Here we go: call ParMETIS
    CALL ParMETIS_F_V3_AdaptiveRepart(vtxdist, xadj%a, adjncy%a, vwgt, vsize, adjwgt, wgtflag, numflag, &
                                      ncon, nparts, tpwgts, ubvec, itr, options, edgecut, part, comm)
                                      
                                      
#ifdef FD4_BALANCE_METIS_DEBUG    
    write(100+rank,*) 'part ',lbound(part,1),' - ',ubound(part,1)
    write(100+rank,*) part
#endif    
    
    allocate(part_global(0:all_blocks_len-1))
    
    do i=0,nprocs-1
      domain%mpi%recvcnt(i) = vtxdist(i+1) - vtxdist(i)
    end do

    ! send local part vector to all other processes
    call MPI_Allgatherv (part, xadj%pos-1, MPI_I4K, part_global, domain%mpi%recvcnt, vtxdist, MPI_I4K, comm, err )
    err = err - MPI_SUCCESS
    FD4_L3_ASSERT(err/=0,'MPI_Allgatherv failed')


#ifdef FD4_BALANCE_METIS_DEBUG    
    write(100+rank,*) 'part_global ',lbound(part_global,1),' - ',ubound(part_global,1)
    write(100+rank,*) part_global
#endif    


    ! Emergency block migration
    ! - if number of blocks >= number of processes:
    !   - find processes without blocks
    !   - move blocks to them
    if(all_blocks_len>=nprocs) then
      vtxcounter = 0
      ! count number of blocks of each process
      do i=0,all_blocks_len-1
        vtxcounter(part_global(i)) = vtxcounter(part_global(i)) + 1
      end do
      ! find processes without blocks
      do ia=0,nprocs-1
        if(vtxcounter(ia)==0) then
          ! this process has no block
          ! find the process with the most blocks
          idx = MAXLOC(vtxcounter(0:nprocs-1))
          idx(1) = idx(1) - 1  ! needed because of lbound = 0, maxloc assumes lbound = 1
#if ( FD4_VERBOSE_LEVEL >= 2 )
          if(rank==0) then
            write(fd4_out,'(A,A)') FD4_LINEHD, &
               'Emergency block migration from '//i2s(idx(1))//' to '//i2s(ia)
            write(fd4_out,'(A,A)') FD4_LINEHD, &
               '          block counts:        '//i2s(vtxcounter(idx(1)))//'    '//i2s(vtxcounter(ia))
          end if
#endif
          vtxcounter(idx(1)) = vtxcounter(idx(1)) - 1
          do i=0,all_blocks_len-1
            ! assign the first block of this process to the process without blocks
            if(part_global(i) == idx(1)) then
              part_global(i) = ia
              exit
            end if
          end do
          !vtxcounter(ia) = 1 ! don't need this
        end if
      end do
    end if



    ! write position of my blocks to local_blocks
    ! and fill domain%blkmap
    domain%blkmap(:,:,:,domain%blkmap_now) = -1
    local_blocks_len = 0
    do i=1,all_blocks_len
      pos = all_blocks(:,i)
      crank = part_global(blkidx(pos(1),pos(2),pos(3)))
      domain%blkmap( pos(1), pos(2), pos(3), domain%blkmap_now ) = crank
      if(rank == crank) then
        local_blocks_len = local_blocks_len + 1
        local_blocks(:,local_blocks_len) = pos(:)
      end if
    end do

#ifdef FD4_BALANCE_METIS_DEBUG    
    write(100+rank,*) 'all_blocks ',1,' - ',all_blocks_len
    do i=1,all_blocks_len
      write(100+rank,*) all_blocks(:,i)
    end do
    write(100+rank,*) 'local_blocks ',1,' - ',local_blocks_len
    do i=1,local_blocks_len
      write(100+rank,*) local_blocks(:,i)
    end do
#endif

    ! apply periodic boundary on blkmap (which is allocated with a boundary to make life easier later on)
    pos = domain%blnum
    if(domain%periodic(1)) then
      domain%blkmap(0,1:pos(2),1:pos(3),domain%blkmap_now) = domain%blkmap(pos(1),1:pos(2),1:pos(3),domain%blkmap_now)
      domain%blkmap(pos(1)+1,1:pos(2),1:pos(3),domain%blkmap_now) = domain%blkmap(1,1:pos(2),1:pos(3),domain%blkmap_now)
    end if
    if(domain%periodic(2)) then
      domain%blkmap(1:pos(1),0,1:pos(3),domain%blkmap_now) = domain%blkmap(1:pos(1),pos(2),1:pos(3),domain%blkmap_now)
      domain%blkmap(1:pos(1),pos(2)+1,1:pos(3),domain%blkmap_now) = domain%blkmap(1:pos(1),1,1:pos(3),domain%blkmap_now)
    end if
    if(domain%periodic(3)) then
      domain%blkmap(1:pos(1),1:pos(2),0,domain%blkmap_now) = domain%blkmap(1:pos(1),1:pos(2),pos(3),domain%blkmap_now)
      domain%blkmap(1:pos(1),1:pos(2),pos(3)+1,domain%blkmap_now) = domain%blkmap(1:pos(1),1:pos(2),1,domain%blkmap_now)
    end if      

    call stack_delete(xadj)
    call stack_delete(adjncy)
    
    
#if ( FD4_VERBOSE_LEVEL >= 4 )
    write(fd4_out,'(A,A,I6)') FD4_LINEHD,'new partition: ',local_blocks_len
#endif
  
  
  end subroutine fd4_part_metis













  !! Create a new partition using ParMETIS. Only a subset of ranks calls ParMETIS.
  !!
  !! *WORK IN PROGRESS*
  !!
  !! *TODO*:
  !! - guess subset from architecture (node), but limit to, let's say, 1/10 of total number of procs (big SMPs)
  !! - create m subsets of size n, with n*m=nprocs. let each subset compute partitioning, hoping that they all
  !!   create the same result.
  !! - what happens if subsets are of different size? equal results then?
  !!
  !!
  subroutine fd4_part_metis_subset(domain, all_blocks, all_blocks_len, local_blocks, local_blocks_len, ubvec, itr, subset)
    type(fd4_domain), intent(inout) :: domain                         !! the domain
    integer(i_k)    , intent(in)    :: all_blocks_len                 !! length of $all_blocks(3,:)
    integer(i_k)    , intent(in)    :: all_blocks(3,all_blocks_len)   !! array of positions of the blocks in the new partion
    integer(i_k)    , intent(inout) :: local_blocks(:,:)              !! array of new local partition (differs from rank to rank)
    integer(i_k)    , intent(out)   :: local_blocks_len               !! length of $local_blocks(3,:)
    real(r4k)       , intent(in)    :: ubvec(1)                       !! ParMetis ubvec parameter
    real(r4k)       , intent(in)    :: itr                            !! ParMetis itr parameter
    integer(i_k)    , intent(in)    :: subset                         !! use only $subset processes for calculation (<2...off)
    
    integer :: rank, nprocs, comm
    integer(i_k), allocatable :: blkidx(:,:,:), blknew(:,:,:)
    integer(i_k) :: i, pos(3), npos(3), crank, dim, dir, iv, ia, err, idx(1)
    
    ! ParMetis parameters
    integer(i4k), allocatable :: vtxdist(:), part(:), vtxcounter(:), vwgt(:), adjwgt(:), vsize(:), part_global(:)
    type(stack) :: xadj, adjncy
    integer(i4k) :: wgtflag, numflag, ncon, nparts, options(0:3), edgecut
    real(r4k), allocatable :: tpwgts(:)
    
    ! communicator stuff
    integer :: comm_subset
    integer :: color, nprocs_subset
    
    
    rank   = domain%mpi%rank
    nprocs = domain%mpi%nprocs
    comm   = domain%mpi%comm

#if ( FD4_VERBOSE_LEVEL >= 1 )
      if(nprocs<2) then
        FD4_L1_ERROR('ParMetis load balancing requires at least 2 processes!')
      end if
#endif    

    !TODO: alternatively a clever guess from architecture once in initialization?
    nprocs_subset = MIN(subset, nprocs)
    
    ! create a new communicator consisting only of the first nprocs_subset ranks
    color = MPI_UNDEFINED
    if(rank<nprocs_subset) color=0
    call MPI_Comm_split(comm,color,rank,comm_subset,err)
  
    
    ! back-up blkmap
    ! we modify blkmap_now in the following before writing the final partition to it
    domain%blkmap_last = domain%blkmap_now
    domain%blkmap_now  = 3 - domain%blkmap_now
    domain%blkmap(:,:,:,domain%blkmap_now) = domain%blkmap(:,:,:,domain%blkmap_last)

    ! mapping: 3D block coordinate -> index for parmetis
    allocate(blkidx(0:domain%blnum(1)+1,0:domain%blnum(2)+1,0:domain%blnum(3)+1))
    ! which block is on which rank?
    allocate(blknew(0:domain%blnum(1)+1,0:domain%blnum(2)+1,0:domain%blnum(3)+1))
    
    
    allocate(vtxdist(0:nprocs_subset))
    vtxdist = 0
    allocate(vtxcounter(0:nprocs))

    call stack_create(xadj,domain%blocks%length+10)
    call stack_create(adjncy,6*domain%blocks%length)
    
    
    ! create a temporary block map (blknew)
    !  - the subset-ranks devide the blocks equally
    !  - removed blocks are not considered
    ! create vtxdist for parmetis
    blknew = -1
    do i=1,all_blocks_len
      pos = all_blocks(:,i)
      crank = (nprocs_subset*(i-1)) / all_blocks_len
      vtxdist(crank+1) = vtxdist(crank+1) + 1
      blknew(pos(1),pos(2),pos(3)) = crank
    end do

    ! sum vtxdist up
    do i=2,nprocs_subset
      vtxdist(i) = vtxdist(i) + vtxdist(i-1)
    end do
    vtxcounter(0:nprocs_subset)=vtxdist(0:nprocs_subset)


    ! create blkidx, maps 3D block coordinates to rank-consecutive IDs used for parmetis
    blkidx = -1
    do i=1,all_blocks_len
      pos = all_blocks(:,i)
      crank = blknew(pos(1),pos(2),pos(3))
      blkidx(pos(1),pos(2),pos(3)) = vtxcounter(crank)
      vtxcounter(crank) = vtxcounter(crank) + 1
    end do    

    ! apply periodic boundary on blknew and blkidx
    pos = domain%blnum
    if(domain%periodic(1)) then
      blkidx(0,1:pos(2),1:pos(3)) = blkidx(pos(1),1:pos(2),1:pos(3))
      blkidx(pos(1)+1,1:pos(2),1:pos(3)) = blkidx(1,1:pos(2),1:pos(3))
      blknew(0,1:pos(2),1:pos(3)) = blknew(pos(1),1:pos(2),1:pos(3))
      blknew(pos(1)+1,1:pos(2),1:pos(3)) = blknew(1,1:pos(2),1:pos(3))
    end if
    if(domain%periodic(2)) then
      blkidx(1:pos(1),0,1:pos(3)) = blkidx(1:pos(1),pos(2),1:pos(3))
      blkidx(1:pos(1),pos(2)+1,1:pos(3)) = blkidx(1:pos(1),1,1:pos(3))
      blknew(1:pos(1),0,1:pos(3)) = blknew(1:pos(1),pos(2),1:pos(3))
      blknew(1:pos(1),pos(2)+1,1:pos(3)) = blknew(1:pos(1),1,1:pos(3))
    end if
    if(domain%periodic(3)) then
      blkidx(1:pos(1),1:pos(2),0) = blkidx(1:pos(1),1:pos(2),pos(3))
      blkidx(1:pos(1),1:pos(2),pos(3)+1) = blkidx(1:pos(1),1:pos(2),1)
      blknew(1:pos(1),1:pos(2),0) = blknew(1:pos(1),1:pos(2),pos(3))
      blknew(1:pos(1),1:pos(2),pos(3)+1) = blknew(1:pos(1),1:pos(2),1)
    end if      
    
#ifdef FD4_BALANCE_METIS_DEBUG    
    do iv=1,domain%blnum(3)
      do ia=1,domain%blnum(2)
        write(100+rank,*) blkidx(1:domain%blnum(1),ia,iv)
      end do
      write(100+rank,*) '---'
    end do
    do iv=1,domain%blnum(3)
      do ia=1,domain%blnum(2)
        write(100+rank,*) blknew(1:domain%blnum(1),ia,iv)
      end do
      write(100+rank,*) '---'
    end do
#endif    


    if(color /= MPI_UNDEFINED) then
    
      ! Create xadj and adjncy for parmetis
      !
      call stack_push(xadj,0)    
      do i=1,all_blocks_len
        pos = all_blocks(:,i)
        if( blknew(pos(1),pos(2),pos(3)) == rank ) then
          do dim=1,3
            do dir=1,2
              npos(:) = pos(:)
              npos(dim) = npos(dim) + fd4_incdec(dir)
              ! put all neighbors to adjcny
              if(blkidx(npos(1),npos(2),npos(3))>=0) then
                call stack_push(adjncy,blkidx(npos(1),npos(2),npos(3)))
              end if
            end do
          end do            
          call stack_push(xadj,adjncy%pos)
        end if
      end do


#ifdef FD4_BALANCE_METIS_DEBUG    
      write(100+rank,*) 'xadj ',0,' - ',xadj%pos-1
      write(100+rank,*) xadj%a(1:xadj%pos)
      write(100+rank,*) 'adjncy ',0,' - ',adjncy%pos-1
      write(100+rank,*) adjncy%a(1:adjncy%pos)
      write(100+rank,*) 'vtxdist ',lbound(vtxdist),' - ',ubound(vtxdist)
      write(100+rank,*) vtxdist
#endif    


      allocate(vwgt(0:xadj%pos-2))
      allocate(vsize(0:xadj%pos-2))
      allocate(adjwgt(0:adjncy%pos-1))
      vwgt=0 
      vsize=0
      adjwgt=0

      ! set vwgt, vsize, and adjwgt
      !
      !TODO: consider real block sizes when calculating vwgt, vsize, and adjwgt
      !TODO:  -> this is easy if all blocks are of equal size
      !TODO: consider real workload of block for vwgt
      do i=1,all_blocks_len
        pos = all_blocks(:,i)
        if( blknew(pos(1),pos(2),pos(3)) == rank ) then
          iv = blkidx(pos(1),pos(2),pos(3)) - vtxdist(rank)
          FD4_L3_ASSERT(iv<0,'inconsistency in calc. of adjwgt')
          FD4_L3_ASSERT(iv>xadj%pos-2,'inconsistency in calc. of adjwgt')
          vwgt(iv) = 1        ! comp. weight of the block
          if(domain%blkmap(pos(1),pos(2),pos(3),domain%blkmap_now) >= 0) vsize(iv) = 1  ! redistribution costs of the block
          ia=xadj%a(iv+1)
          do dim=1,3
            do dir=1,2
              npos(:) = pos(:)
              npos(dim) = npos(dim) + fd4_incdec(dir)
              if(blknew(npos(1),npos(2),npos(3))>=0) then
                adjwgt(ia)=1
                ia=ia+1
              end if
            end do
          end do
          FD4_L3_ASSERT(xadj%a(iv+2)/=ia,'inconsistency in calc. of adjwgt')
        end if      
      end do    



#ifdef FD4_BALANCE_METIS_DEBUG    
      write(100+rank,*) 'vwgt ',lbound(vwgt,1),' - ',ubound(vwgt,1)
      write(100+rank,*) vwgt
      write(100+rank,*) 'vsize ',lbound(vsize,1),' - ',ubound(vsize,1)
      write(100+rank,*) vsize
      write(100+rank,*) 'adjwgt ',lbound(adjwgt,1),' - ',ubound(adjwgt,1)
      write(100+rank,*) adjwgt
#endif

      wgtflag = 3
      numflag = 0 ! we are using C-Style numbering!
      ncon = 1
      nparts = nprocs
      allocate(tpwgts(nprocs))
      tpwgts = 1.0_r4k/REAL(nparts,r4k)
      !options(0:3) = (/1,127,15,1/)
      !options(0:3) = 0
      options(0:3) = (/1,0,15,2/)
      allocate(part(0:xadj%pos-2))
      part = 0

      ! fill part with current partitioning, rank 0 takes the new blocks 
      do i=1,all_blocks_len
        crank = (nprocs_subset*(i-1)) / all_blocks_len
        if(rank==crank) then
          pos = all_blocks(:,i)
          iv = blkidx(pos(1),pos(2),pos(3)) - vtxdist(crank)
          !print *,rank,pos,iv
          part(iv) = MAX(0,domain%blkmap(pos(1),pos(2),pos(3),domain%blkmap_now))
        end if
      end do    



#ifdef FD4_PART_METIS_DUMP
      if(fd4_part_metis_dump>=1) then
        if(rank==0) write(fd4_out,'(A,A)') FD4_LINEHD,'dumping ParMetis parameters...'
        !! dump parmetis parameters to file
        write(1000+rank,*) lbound(vtxdist,1),ubound(vtxdist,1),'vtxdist'
        write(1000+rank,*) vtxdist
        write(1000+rank,*) 0,xadj%pos-1,'xadj'
        write(1000+rank,*) xadj%a(1:xadj%pos)
        write(1000+rank,*) 0,adjncy%pos-1,'adjncy'
        write(1000+rank,*) adjncy%a(1:adjncy%pos)
        write(1000+rank,*) lbound(vwgt,1),ubound(vwgt,1),'vwgt'
        write(1000+rank,*) vwgt
        write(1000+rank,*) lbound(vsize,1),ubound(vsize,1),'vsize'
        write(1000+rank,*) vsize
        write(1000+rank,*) lbound(adjwgt,1),ubound(adjwgt,1),'adjwgt'
        write(1000+rank,*) adjwgt
        write(1000+rank,*) wgtflag
        write(1000+rank,*) numflag
        write(1000+rank,*) ncon
        write(1000+rank,*) nparts
        write(1000+rank,*) lbound(tpwgts,1),ubound(tpwgts,1),'tpwgts'
        write(1000+rank,*) tpwgts
        write(1000+rank,*) lbound(ubvec,1),ubound(ubvec,1),'ubvec'
        write(1000+rank,*) ubvec
        write(1000+rank,*) itr
        write(1000+rank,*) lbound(options,1),ubound(options,1),'options'
        write(1000+rank,*) options
        write(1000+rank,*) lbound(part,1),ubound(part,1),'part'
        write(1000+rank,*) part
        CALL MPI_Barrier(comm_subset,err)
        if(rank==0) write(fd4_out,'(A,A)') FD4_LINEHD,'ParMetis parameters dumped.'
        !if(fd4_part_metis_dump==2) call MPI_Abort(MPI_COMM_WORLD,1,err)
        if(fd4_part_metis_dump==2) stop
        fd4_part_metis_dump = 0
      end if
#endif


#if ( FD4_VERBOSE_LEVEL >= 4 )
      if(rank==0) then
        FD4_L4_INFO('calling ParMetis decoupled on '//i2s(nprocs_subset)//' ranks')
      end if
#endif

      ! Here we go: call ParMETIS
      CALL ParMETIS_F_V3_AdaptiveRepart(vtxdist, xadj%a, adjncy%a, vwgt, vsize, adjwgt, wgtflag, numflag, &
                                        ncon, nparts, tpwgts, ubvec, itr, options, edgecut, part, comm_subset)


#ifdef FD4_BALANCE_METIS_DEBUG    
      write(100+rank,*) 'part ',lbound(part,1),' - ',ubound(part,1)
      write(100+rank,*) part
#endif    


      allocate(part_global(0:all_blocks_len-1))

      do i=0,nprocs_subset-1
        domain%mpi%recvcnt(i) = vtxdist(i+1) - vtxdist(i)
      end do

      ! send local part vector to all other processes
      !TODO: MPI_Gatherv would be enough here, since we bcast the stuff to all procs later
      call MPI_Gatherv (part, xadj%pos-1, MPI_I4K, part_global, domain%mpi%recvcnt, vtxdist, MPI_I4K, 0, comm_subset, err )
      !call MPI_Allgatherv (part, xadj%pos-1, MPI_I4K, part_global, domain%mpi%recvcnt, vtxdist, MPI_I4K, comm_subset, err )
      err = err - MPI_SUCCESS
      FD4_L3_ASSERT(err/=0,'MPI_Gatherv failed')

      call MPI_Comm_free(comm_subset,err)


    else    ! color /= MPI_UNDEFINED
      
      ! processes that do not call parmetis
      allocate(part_global(0:all_blocks_len-1))
      
    end if  ! color /= MPI_UNDEFINED
    
    call MPI_Bcast(part_global,all_blocks_len,MPI_I4K,0,comm,err)
    

#ifdef FD4_BALANCE_METIS_DEBUG    
    write(100+rank,*) 'part_global ',lbound(part_global,1),' - ',ubound(part_global,1)
    write(100+rank,*) part_global
#endif    


    ! Emergency block migration
    ! - if number of blocks >= number of processes:
    !   - find processes without blocks
    !   - move blocks to them
    if(all_blocks_len>=nprocs) then
      vtxcounter = 0
      ! count number of blocks of each process
      do i=0,all_blocks_len-1
        vtxcounter(part_global(i)) = vtxcounter(part_global(i)) + 1
      end do
      ! find processes without blocks
      do ia=0,nprocs-1
        if(vtxcounter(ia)==0) then
          ! this process has no block
          ! find the process with the most blocks
          idx = MAXLOC(vtxcounter(0:nprocs-1))
          idx(1) = idx(1) - 1  ! needed because of lbound = 0, maxloc assumes lbound = 1
#if ( FD4_VERBOSE_LEVEL >= 2 )
          if(rank==0) then
            write(fd4_out,'(A,A)') FD4_LINEHD, &
               'Emergency block migration from '//i2s(idx(1))//' to '//i2s(ia)
            write(fd4_out,'(A,A)') FD4_LINEHD, &
               '          block counts:        '//i2s(vtxcounter(idx(1)))//'    '//i2s(vtxcounter(ia))
          end if
#endif
          vtxcounter(idx(1)) = vtxcounter(idx(1)) - 1
          do i=0,all_blocks_len-1
            ! assign the first block of this process to the process without blocks
            if(part_global(i) == idx(1)) then
              part_global(i) = ia
              exit
            end if
          end do
          !vtxcounter(ia) = 1 ! don't need this
        end if
      end do
    end if



    ! write position of my blocks to local_blocks
    ! and fill domain%blkmap
    domain%blkmap(:,:,:,domain%blkmap_now) = -1
    local_blocks_len = 0
    do i=1,all_blocks_len
      pos = all_blocks(:,i)
      crank = part_global(blkidx(pos(1),pos(2),pos(3)))
      domain%blkmap( pos(1), pos(2), pos(3), domain%blkmap_now ) = crank
      if(rank == crank) then
        local_blocks_len = local_blocks_len + 1
        local_blocks(:,local_blocks_len) = pos(:)
      end if
    end do

#ifdef FD4_BALANCE_METIS_DEBUG    
    write(100+rank,*) 'all_blocks ',1,' - ',all_blocks_len
    do i=1,all_blocks_len
      write(100+rank,*) all_blocks(:,i)
    end do
    write(100+rank,*) 'local_blocks ',1,' - ',local_blocks_len
    do i=1,local_blocks_len
      write(100+rank,*) local_blocks(:,i)
    end do
#endif


    ! apply periodic boundary on blkmap (which is allocated with a boundary to make life easier later on)
    pos = domain%blnum
    if(domain%periodic(1)) then
      domain%blkmap(0,1:pos(2),1:pos(3),domain%blkmap_now) = domain%blkmap(pos(1),1:pos(2),1:pos(3),domain%blkmap_now)
      domain%blkmap(pos(1)+1,1:pos(2),1:pos(3),domain%blkmap_now) = domain%blkmap(1,1:pos(2),1:pos(3),domain%blkmap_now)
    end if
    if(domain%periodic(2)) then
      domain%blkmap(1:pos(1),0,1:pos(3),domain%blkmap_now) = domain%blkmap(1:pos(1),pos(2),1:pos(3),domain%blkmap_now)
      domain%blkmap(1:pos(1),pos(2)+1,1:pos(3),domain%blkmap_now) = domain%blkmap(1:pos(1),1,1:pos(3),domain%blkmap_now)
    end if
    if(domain%periodic(3)) then
      domain%blkmap(1:pos(1),1:pos(2),0,domain%blkmap_now) = domain%blkmap(1:pos(1),1:pos(2),pos(3),domain%blkmap_now)
      domain%blkmap(1:pos(1),1:pos(2),pos(3)+1,domain%blkmap_now) = domain%blkmap(1:pos(1),1:pos(2),1,domain%blkmap_now)
    end if      

    call stack_delete(xadj)
    call stack_delete(adjncy)
    
    
#if ( FD4_VERBOSE_LEVEL >= 4 )
    write(fd4_out,'(A,A,I6)') FD4_LINEHD,'new partition: ',local_blocks_len
#endif
  
  
  end subroutine fd4_part_metis_subset




end module fd4_part_metis_mod
