! This file is part of FD4: Four-Dimensional Distributed Dynamic Data structures
!  
! Copyright 2010 Matthias Lieber, ZIH, TU Dresden, Germany
! matthias.lieber@tu-dresden.de
!  
! FD4 is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!  
! FD4 is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!  
! You should have received a copy of the GNU General Public License
! along with FD4.  If not, see <http://www.gnu.org/licenses/>.

!--------------------------------------------------------------!
!                                                              !
!         Recursive Coordinate Bisection partitioning          !
!                                                              !
!--------------------------------------------------------------!

#include "fd4.h"

!! Recursive Coordinate Bisection partitioning of an $fd4_domain.
!!
!! *CAUTION*: Don't use! Has not been tested for large domains or high process counts.
!!
!! @see fd4_domain_mod
!! @author Matthias Lieber
module fd4_part_rcb_mod

  use fd4_globaldef_mod
  use fd4_domain_mod
  use fd4_iter_mod
  use fd4_mpi_mod
  use rbtree_fd4_block_mod
  
  USE_MPI
  
  implicit none
  private
  
  INCLUDE_MPIFH
  
  ! public methods
  public :: fd4_part_rcb_weight

  
contains  

  include 'framework/fd4_domain.incf'
  include 'util/math.incf'


  !! Create a new partition using block position. Considers block weights from domain%balance%blkweights.
  !!
  !! Only works with processor counts which are powers of 2.
  !!
  !! *TODO*:
  !!  - set output argument $balance properly
  !!  - enable arbitrary processor counts
  !!
  subroutine fd4_part_rcb_weight(domain, all_blocks, all_blocks_len, local_blocks, local_blocks_len,  &
                                 balance, err)
    type(fd4_domain), intent(inout) :: domain                         !! the domain
    integer(i_k)    , intent(in)    :: all_blocks_len                 !! length of $all_blocks(3,:)
    integer(i_k)    , intent(in)    :: all_blocks(3,all_blocks_len)   !! array of positions of the blocks in the new partion
    integer(i_k)    , intent(inout) :: local_blocks(:,:)              !! array of new local partition (differs from rank to rank)
    integer(i_k)    , intent(out)   :: local_blocks_len               !! length of $local_blocks(3,:)
    real(r4k)       , intent(out)   :: balance                        !! estimated balance of new partitioning (not implemented yet)
    integer(i_k)    , intent(out)   :: err                            !! error status: 0...ok
    
    integer :: rank, nprocs
    integer(i_k) :: steps
    integer(i_k) :: i, j, k, pos(3), bnd(3,2)
    
    err = 0
    rank   = domain%mpi%rank
    nprocs = domain%mpi%nprocs
    
    steps = log2i(nprocs)
    FD4_L1_ASSERT(2**steps/=nprocs,'nproc is no power of 2')
    
    ! collect the block weights to domain%balance%blkweights
    call fd4_part_rcb_exch_weights(domain, all_blocks, all_blocks_len, err)
    FD4_L1_ASSERT(err/=0,'fd4_part_rcb_exch_weights failed')
    
    domain%blkmap_last = domain%blkmap_now
    domain%blkmap_now  = 3 - domain%blkmap_now

    if(domain%varinfo%vthres_mode) then
      ! adaptive block mode: set block map to -1 in all not required blocks
      domain%blkmap(:,:,:,domain%blkmap_now) = -1
      do i=1,all_blocks_len
        domain%blkmap(all_blocks(1,i),all_blocks(2,i),all_blocks(3,i),domain%blkmap_now) = 0
      end do
    end if

    ! call recursive routine which performs RCB partitioning
    ! modifies domain%blkmap
    bnd(1:3,1) = 1
    bnd(1:3,2) = domain%blnum(1:3)
    call fd4_part_rcb_bisect(domain,bnd,(/0,nprocs-1/))

    ! get local_blocks from new blkmap
    local_blocks_len = 0
    do k=1,domain%blnum(3)
      do j=1,domain%blnum(2)
        do i=1,domain%blnum(1)
          if( rank == domain%blkmap(i,j,k,domain%blkmap_now) ) then
            local_blocks_len = local_blocks_len + 1
            local_blocks(1,local_blocks_len) = i
            local_blocks(2,local_blocks_len) = j
            local_blocks(3,local_blocks_len) = k
          end if
        end do
      end do
    end do
    
    !print *,rank,' : ',local_blocks_len, local_blocks(:,local_blocks_len)
    
    !! TODO: fill balance
    balance = -2

    
    ! apply periodic boundary on blkmap (which is allocated with a boundary to make life easier later on)
    pos = domain%blnum
    if(domain%periodic(1)) then
      domain%blkmap(0,1:pos(2),1:pos(3),domain%blkmap_now) = domain%blkmap(pos(1),1:pos(2),1:pos(3),domain%blkmap_now)
      domain%blkmap(pos(1)+1,1:pos(2),1:pos(3),domain%blkmap_now) = domain%blkmap(1,1:pos(2),1:pos(3),domain%blkmap_now)
    end if
    if(domain%periodic(2)) then
      domain%blkmap(1:pos(1),0,1:pos(3),domain%blkmap_now) = domain%blkmap(1:pos(1),pos(2),1:pos(3),domain%blkmap_now)
      domain%blkmap(1:pos(1),pos(2)+1,1:pos(3),domain%blkmap_now) = domain%blkmap(1:pos(1),1,1:pos(3),domain%blkmap_now)
    end if
    if(domain%periodic(3)) then
      domain%blkmap(1:pos(1),1:pos(2),0,domain%blkmap_now) = domain%blkmap(1:pos(1),1:pos(2),pos(3),domain%blkmap_now)
      domain%blkmap(1:pos(1),1:pos(2),pos(3)+1,domain%blkmap_now) = domain%blkmap(1:pos(1),1:pos(2),1,domain%blkmap_now)
    end if      
    
    
#if ( FD4_VERBOSE_LEVEL >= 4 )
    write(fd4_out,'(A,A,I6,E13.4)') FD4_LINEHD,'new partition: ',local_blocks_len, domain%balance%localweight
#endif
   
  
  end subroutine fd4_part_rcb_weight


  !! Recursive bisetion function.
  !!
  !! Internal routine for $fd4_part_rcb_weight.
  recursive subroutine fd4_part_rcb_bisect(domain, bnd, ranks)
    type(fd4_domain), intent(inout) :: domain      !! the domain
    integer(i_k)    , intent(in)    :: bnd(3,2)    !! current subset
    integer(i_k)    , intent(in)    :: ranks(2)    !! rank interval to assign
    
    integer :: dim_len(3), dim, i, cut, bnd1(3,2), bnd2(3,2)
    real(r4k) :: targetwgt, curwgt, lastcurwgt
    
    dim_len = bnd(:,2) - bnd(:,1) + 1
    dim = maxloc(dim_len,1)

    ! only one rank left or only one block left?
    if(ranks(1)==ranks(2) .or. dim_len(dim)==1) then
      ! fill blkmap
      if(domain%varinfo%vthres_mode) then
        where(domain%blkmap(bnd(1,1):bnd(1,2),bnd(2,1):bnd(2,2),bnd(3,1):bnd(3,2),domain%blkmap_now)>=0)
          domain%blkmap(bnd(1,1):bnd(1,2),bnd(2,1):bnd(2,2),bnd(3,1):bnd(3,2),domain%blkmap_now) = ranks(1)
        end where
      else
        domain%blkmap(bnd(1,1):bnd(1,2),bnd(2,1):bnd(2,2),bnd(3,1):bnd(3,2),domain%blkmap_now) = ranks(1)
      end if
      return
    end if
    
    targetwgt = SUM( domain%balance%blkweights(bnd(1,1):bnd(1,2),bnd(2,1):bnd(2,2),bnd(3,1):bnd(3,2)) ) / 2.0
    lastcurwgt = 0
    curwgt = 0
    bnd1=bnd
    bnd2=bnd
    cut=(bnd(dim,1)+bnd(dim,2))/2

    if(dim==1) then
      do i=bnd(dim,1),bnd(dim,2)
        lastcurwgt = curwgt
        curwgt = curwgt + SUM( domain%balance%blkweights(i,bnd(2,1):bnd(2,2),bnd(3,1):bnd(3,2)) )
        if(curwgt>targetwgt) then
          cut=i
          exit
        end if
      end do
    end if
    
    if(dim==2) then
      do i=bnd(dim,1),bnd(dim,2)
        lastcurwgt = curwgt
        curwgt = curwgt + SUM( domain%balance%blkweights(bnd(1,1):bnd(1,2),i,bnd(3,1):bnd(3,2)) )
        if(curwgt>targetwgt) then
          cut=i
          exit
        end if
      end do
    end if

    if(dim==3) then
      do i=bnd(dim,1),bnd(dim,2)
        lastcurwgt = curwgt
        curwgt = curwgt + SUM( domain%balance%blkweights(bnd(1,1):bnd(1,2),bnd(2,1):bnd(2,2),i) )
        if(curwgt>targetwgt) then
          cut=i
          exit
        end if
      end do
    end if

    if( (targetwgt - lastcurwgt) < (curwgt - targetwgt) ) then
      cut = cut - 1
    end if
    
    bnd1(dim,2) = cut
    bnd2(dim,1) = cut + 1
    call fd4_part_rcb_bisect(domain, bnd1, (/ ranks(1)          , (SUM(ranks)-1)/2 /) )
    call fd4_part_rcb_bisect(domain, bnd2, (/ (SUM(ranks)-1)/2+1, ranks(2)         /) )

  end subroutine fd4_part_rcb_bisect


  !! Exchange the weight of my blocks with all other processes.
  !!
  !! Updates domain%balance%blkweights.
  !! This is a copy from $fd4_part_sfc with modifications (argument all_blocks).
  !!
  !! @see fd4_part_sfc_mod
  !! Internal routine.
  subroutine fd4_part_rcb_exch_weights(domain, all_blocks, all_blocks_len, err)
    type(fd4_domain), intent(inout), target :: domain          !! the domain
    integer(i_k)    , intent(in)            :: all_blocks_len  !! length of $all_blocks(3,:)
    integer(i_k)    , intent(in)            :: all_blocks(3,all_blocks_len) !! array of positions of the blocks in the new partion
    integer(i_k),     intent(out)           :: err             !! error status: 0...ok
    
    type(fd4_mpi), pointer :: mpi
    integer(i_k) :: crank, nblocks, k, j, i
    type(fd4_iter) :: iter

    mpi => domain%mpi
    
    mpi%recvcnt(:) = 0
    nblocks = 0
    err = -1
    
    ! loop over blkmap to build recv counts for MPI_Allgatherv
    ! (number of blocks for each process)
    do k=1,domain%blnum(3)
      do j=1,domain%blnum(2)
        do i=1,domain%blnum(1)
          crank = domain%blkmap(i,j,k,domain%blkmap_now)
          if( crank >= 0 ) then
            mpi%recvcnt(crank) = mpi%recvcnt(crank) + 1
          end if
        end do
      end do
    end do

    FD4_L3_ASSERT(mpi%recvcnt(mpi%rank)/=domain%blocks%length,'nblocks not as expected')
    
    ! build the recv displacements
    mpi%recvdis(0) = 0
    do crank=1,mpi%nprocs-1
      mpi%recvdis(crank) = mpi%recvcnt(crank-1) + mpi%recvdis(crank-1)
    end do

    nblocks = mpi%recvdis(mpi%nprocs-1) + mpi%recvcnt(mpi%nprocs-1)

    ! this may fail, because domain%blockcount may be inconsistent
    ! after calling fd4_util_put_array
    !FD4_L3_ASSERT(nblocks/=domain%blockcount,'nblocks not as expected')
    
    ! allocate arrays for sending and receiving the block weights
    if( size(mpi%recvreal,1) < nblocks ) then
      deallocate(mpi%recvreal)
      allocate(mpi%recvreal(nblocks))
    end if
    if( size(mpi%sendreal,1) < mpi%recvcnt(mpi%rank) ) then
      deallocate(mpi%sendreal)
      allocate(mpi%sendreal(mpi%recvcnt(mpi%rank)))
    end if
    
    ! get the local block weights and write them to mpi%sendreal
    i = 0
    iter%cur => rbtree_fd4_block_iterator_first(iter%biter, domain%blocks)
    do while(associated(iter%cur))
      i = i + 1
      mpi%sendreal(i) = iter%cur%weight
      iter%cur => rbtree_fd4_block_iterator_next(iter%biter)
    end do
    
    FD4_L3_ASSERT(i/=domain%blocks%length,'nblocks not as expected')
        

    ! exchange the block weights
    call MPI_Allgatherv (mpi%sendreal, mpi%recvcnt(mpi%rank),              MPI_R4K, &
                         mpi%recvreal, mpi%recvcnt,           mpi%recvdis, MPI_R4K, mpi%comm, err )
    err = err - MPI_SUCCESS
    FD4_L1_ASSERT(err/=0,'MPI_Allgatherv failed')
    
    
    if(domain%varinfo%vthres_mode) then
      ! adaptive block mode: reset block weights to force only those blocks weights
      ! to be set, which are required in new partitioning (all_blocks)
      domain%balance%blkweights = 0.0_r4k
      do i=1,all_blocks_len
        domain%balance%blkweights(all_blocks(1,i),all_blocks(2,i),all_blocks(3,i)) = 1.0_r4k
      end do
    end if
    
    
    ! loop over blkmap and fill domain%balance%blkweights
    ! this method assumes the loop order in this loop has the same ordering
    ! as the index used to sort the blocks in domain%blocks
    do k=1,domain%blnum(3)
      do j=1,domain%blnum(2)
        do i=1,domain%blnum(1)
          crank = domain%blkmap(i,j,k,domain%blkmap_now)
          ! in adaptive block mode: only set blkweights when block is required in new partitioning (see above)
          if( (.not. domain%varinfo%vthres_mode) .or. domain%balance%blkweights(i,j,k)>0.0 ) then
            if( crank >= 0 ) then
              domain%balance%blkweights(i,j,k) = mpi%recvreal( mpi%recvdis(crank)+1 )
            else
              ! set default weight
              !TODO: do this only if the block has been removed last load balancing
              !      (or block will be added this time)
              domain%balance%blkweights(i,j,k) = domain%balance%defaultweight  &
               * REAL(   ( domain%d(1)%bl(i+1) - domain%d(1)%bl(i) )           &
                       * ( domain%d(2)%bl(j+1) - domain%d(2)%bl(j) )           &
                       * ( domain%d(3)%bl(k+1) - domain%d(3)%bl(k) ), r4k )
            end if
          end if
          if( crank >= 0 ) mpi%recvdis(crank) = mpi%recvdis(crank) + 1
        end do
      end do
    end do

#if ( FD4_VERBOSE_LEVEL >= 3 )
    do crank=1,mpi%nprocs-1
      if(mpi%recvdis(crank) - mpi%recvdis(crank-1) /= mpi%recvcnt(crank)) then
        FD4_L1_ERROR('inconsistency after loop over blkmap')
      end if
    end do
#endif
    
    err = 0
    
  end subroutine fd4_part_rcb_exch_weights
  



end module fd4_part_rcb_mod
